// Example of usage of OpenCL for Delphi by Michal Pohanka: 2013-07-19.
// http://www.volny.cz/profipohanka

unit FMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, dglOpenGL, CL, CL_GL, ExtCtrls;

const
  c_device_types:array[1..3] of cardinal=(CL_DEVICE_TYPE_CPU, CL_DEVICE_TYPE_GPU, CL_DEVICE_TYPE_ACCELERATOR);
  
type
	// platform+device_type couples
	t_platform_device=record
		platform_id: cl_platform_id; // index for platform_id
		device_type: cl_device_type; // CL_DEVICE_TYPE_CPU; CL_DEVICE_TYPE_GPU; CL_DEVICE_TYPE_ACCELERATOR
	end;

  TFormMain = class(TForm)
    ListOutput: TListBox;
    Panel1: TPanel;
    BtnReadOpenCLCouples: TButton;
    BtnFncNotLoaded: TButton;
    ListCouple: TComboBox;
    Label1: TLabel;
    BtnTestContext: TButton;
    BtnTestProgram: TButton;
    procedure BtnReadOpenCLCouplesClick(Sender: TObject);
    procedure BtnTestContextClick(Sender: TObject);
    procedure BtnTestProgramClick(Sender: TObject);
    procedure BtnFncNotLoadedClick(Sender: TObject);
  private
    errcode_ret: cl_int; // error code returned from api calls
    platform_devices: array of t_platform_device;
  	num_devices_returned: cl_uint;
  	device_ids: array of cl_device_id; // compute device IDs
  	context: cl_context; // compute context

    function convertToString(const Filename: AnsiString): AnsiString;
  public
    { Public-Deklarationen }
  end;

var
  FormMain: TFormMain;

implementation

{$R *.dfm}

uses
  math;

// README:
//
// Things to take care of:
//
// Be sure to use aligned memory, as OpenCL expects it! (Therefore FastMM4
// is included in the dpr-file, newer Delphi versions come with FastMM4 as the
// default memory manager. But be sure that everything you pass to OpenCl is
// aligned on a 16 byte boundary, also for example things you allocate on the
// stack.)

function TFormMain.convertToString(const Filename: AnsiString): AnsiString;
begin
  with TFileStream.Create(IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName))+Filename, fmOpenRead or fmShareDenyWrite) do
  try
    SetLength(Result, Size);
    if Size > 0 then
      Read(Result[1], Size);
  finally
    Free;
  end;
end;


procedure TFormMain.BtnReadOpenCLCouplesClick(Sender: TObject);
var
  i, j: integer;
	num_platforms_returned: cl_uint;
  platform_id: array of cl_platform_id; // platform id
  i_pl: integer;
  i_type: integer;
  device_vendor, device_name: AnsiString;
  returned_size: size_t;
begin
  if not OpenCL_loaded then begin
    ShowMessage('Error: OpenCL.dll not loaded!');
    exit;
  end;

  // Number of platforms
	errcode_ret:=clGetPlatformIDs(0, nil, @num_platforms_returned);
	if (errcode_ret <> CL_SUCCESS) then begin
		ShowMessage('Error: Failed to get number of platforms!');
		exit;
  end;
  // Connect to a platform
  SetLength(platform_id, num_platforms_returned);
	errcode_ret:=clGetPlatformIDs(num_platforms_returned, @platform_id[0], @num_platforms_returned);
	if (errcode_ret <> CL_SUCCESS) then begin
		ShowMessage('Error: Failed to find platform!');
		exit;
  end;

	// find platform+device_type
	SetLength(platform_devices, 0);
  for i_pl:=0 to num_platforms_returned-1 do for i_type:=1 to 3 do
  	if (CL_SUCCESS=clGetDeviceIDs(platform_id[i_pl], c_device_types[i_type], 0, nil, @num_devices_returned)) and (num_devices_returned>=1)
    then begin
    	SetLength(platform_devices, Length(platform_devices)+1);
		  platform_devices[High(platform_devices)].platform_id:=platform_id[i_pl];
			platform_devices[High(platform_devices)].device_type:=c_device_types[i_type];
    end;
	// list platform+device_type couples
  ListCouple.Clear;
	for i:=0 to High(platform_devices) do begin
		errcode_ret:=clGetDeviceIDs(platform_devices[i].platform_id, platform_devices[i].device_type, 0, nil, @num_devices_returned);
		if (errcode_ret<>CL_SUCCESS) then begin
  		ShowMessage('Error: Failed to create a device group!');
	  	exit;
    end;
		SetLength(device_ids, num_devices_returned);
		errcode_ret:=clGetDeviceIDs(platform_devices[i].platform_id, platform_devices[i].device_type, num_devices_returned, @device_ids[0], @num_devices_returned);
		for j:=0 to num_devices_returned-1 do begin
      SetLength(device_name, 1024);
			clGetDeviceInfo(device_ids[j], CL_DEVICE_NAME, Length(device_name), PAnsiChar(device_name), @returned_size);
      SetLength(device_name, Min(Pos(#0, device_name)-1, returned_size-1));
      SetLength(device_vendor, 1024);
			clGetDeviceInfo(device_ids[j], CL_DEVICE_VENDOR, Length(device_vendor), PAnsiChar(device_vendor), @returned_size);
      SetLength(device_vendor, Min(Pos(#0, device_vendor)-1, returned_size-1));
      ListCouple.Items.Add('Vendor: '+device_vendor+'; Device: '+device_name);
		end;
		SetLength(device_ids, 0);
	end;
  if ListCouple.Items.Count>0 then ListCouple.ItemIndex:=0;
end;

procedure TFormMain.BtnTestContextClick(Sender: TObject);
begin
  if ListCouple.ItemIndex<0 then begin
    ShowMessage('Select couple!');
    exit;
  end;
	// Get compute devices from platform
	errcode_ret:=clGetDeviceIDs(platform_devices[ListCouple.ItemIndex].platform_id, platform_devices[ListCouple.ItemIndex].device_type, 0, nil, @num_devices_returned);
  SetLength(device_ids, num_devices_returned);
  errcode_ret:=clGetDeviceIDs(platform_devices[ListCouple.ItemIndex].platform_id, platform_devices[ListCouple.ItemIndex].device_type, num_devices_returned, @device_ids[0], @num_devices_returned);
  if (errcode_ret<>CL_SUCCESS) then begin
    ShowMessage('Error: Failed to create a device group!');
    exit;
  end;
	// Create a compute context
	context:=clCreateContext(nil, num_devices_returned, @device_ids[0], nil, nil, @errcode_ret);
  if (errcode_ret<>CL_SUCCESS) then begin
    ShowMessage('Error: Failed to create a compute context!!');
    exit;
  end;
  ShowMessage('Context created.');
 	clReleaseContext(context);
end;

procedure TFormMain.BtnTestProgramClick(Sender: TObject);
const
  MYSIZE=2048;
  // Some interesting data for the vectors
  InitialData1: array[0..19] of cl_float = (37.4,50,54,50,56,0,43,43,7.4,71,32,36,16,43,56,100,50,25,15,17);
  InitialData2: array[0..12] of cl_float = (35,51,54.5,58,55,3.2,36,69,27,39,35,40,16);
var
  i: integer;
	HostVector1, HostVector2, HostOutputVector: array[0..MYSIZE-1] of cl_float;
  sourceStr: AnsiString;
  sourceSize: size_t;
  sourcePAnsiChar: PAnsiChar;
  OpenCLProgram: cl_program;
  OpenCLVectorAdd: cl_kernel;
  CommandQueue: cl_command_queue;
  GPUVector1, GPUVector2, GPUOutputVector: cl_mem;
  globalThreads: array[0..0] of size_t;
  //localThreads: array[0..0] of size_t;
  s, error_string: AnsiString;
  returned_size: size_t;
begin
  Cursor:=crHourGlass;
  ////////////////
  // Init Context
  if ListCouple.ItemIndex<0 then begin
    ShowMessage('Select couple!');
    Cursor:=crDefault;
    exit;
  end;
	// Get compute devices from platform
	errcode_ret:=clGetDeviceIDs(platform_devices[ListCouple.ItemIndex].platform_id, platform_devices[ListCouple.ItemIndex].device_type, 0, nil, @num_devices_returned);
  SetLength(device_ids, num_devices_returned);
  errcode_ret:=clGetDeviceIDs(platform_devices[ListCouple.ItemIndex].platform_id, platform_devices[ListCouple.ItemIndex].device_type, num_devices_returned, @device_ids[0], @num_devices_returned);
  if (errcode_ret<>CL_SUCCESS) then begin
    ShowMessage('Error: Failed to create a device group!');
    Cursor:=crDefault;
    exit;
  end;
	// Create a compute context
	context:=clCreateContext(nil, num_devices_returned, @device_ids[0], nil, nil, @errcode_ret);
  if (errcode_ret<>CL_SUCCESS) then begin
    ShowMessage('Error: Failed to create a compute context!!');
    Cursor:=crDefault;
    exit;
  end;
  // End (Init Context)
  ////////////////

	// Initialize with some interesting repeating data
	for i:=0 to MYSIZE-1 do begin
		HostVector1[i]:=InitialData1[i mod 20];
		HostVector2[i]:=InitialData2[i mod 13];
	end;

  // Create OpenCL program with source code
  sourceStr:=convertToString('VectorAdd.cl');
  sourceSize:=Length(sourceStr);
  sourcePAnsiChar:=PAnsiChar(sourceStr);
  OpenCLProgram := clCreateProgramWithSource(context, 1, @sourcePAnsiChar, @sourceSize, @errcode_ret);
	if errcode_ret<>CL_SUCCESS then begin
    ShowMessage('Error: clCreateProgramWithSource failed!');
    clReleaseContext(context);
    Cursor:=crDefault;
    exit;
  end;

	// Build the program (OpenCL JIT compilation)
	if CL_SUCCESS<>clBuildProgram(OpenCLProgram, 0, nil, nil, nil, nil) then begin
    error_string:='Error: clBuildProgram failed! ';
    clGetProgramBuildInfo(OpenCLProgram, device_ids[0], CL_PROGRAM_BUILD_LOG, 0, nil, @returned_size);
    SetLength(s, returned_size+2);
    clGetProgramBuildInfo(OpenCLProgram, device_ids[0], CL_PROGRAM_BUILD_LOG, Length(s), PAnsiChar(s), @returned_size);
    SetLength(s, Min(Pos(#0, s)-1, returned_size-1));
    error_string:=error_string+s;

    ShowMessage(error_string);
    clReleaseProgram(OpenCLProgram);
    clReleaseContext(context);
    Cursor:=crDefault;
    exit;
  end;
	// Create a handle to the compiled OpenCL function (Kernel)
	OpenCLVectorAdd:=clCreateKernel(OpenCLProgram, PAnsiChar('VectorAdd'), nil);

	// Create a command-queue on the first CPU or GPU device
	CommandQueue:=clCreateCommandQueue(context, device_ids[0], 0, nil);

	// Allocate GPU memory for source vectors AND initialize from CPU memory
	GPUVector1:=clCreateBuffer(context, CL_MEM_READ_ONLY or CL_MEM_COPY_HOST_PTR, sizeof(HostVector1[0]) * MYSIZE, @HostVector1[0], nil);
	GPUVector2:=clCreateBuffer(context, CL_MEM_READ_ONLY or CL_MEM_COPY_HOST_PTR, sizeof(HostVector2[0]) * MYSIZE, @HostVector2[0], nil);
	// Allocate output memory on GPU
	GPUOutputVector:=clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(HostOutputVector[0]) * MYSIZE, nil, nil);
	// In the next step we associate the GPU memory with the Kernel arguments
	clSetKernelArg(OpenCLVectorAdd, 0, sizeof(cl_mem), @GPUOutputVector);
	clSetKernelArg(OpenCLVectorAdd, 1, sizeof(cl_mem), @GPUVector1);
	clSetKernelArg(OpenCLVectorAdd, 2, sizeof(cl_mem), @GPUVector2);
	// Launch the Kernel on the GPU
  globalThreads[0]:=MYSIZE;
  //localThreads[0]:=1;
	clEnqueueNDRangeKernel(CommandQueue, OpenCLVectorAdd, 1, nil, @globalThreads, nil, 0, nil, nil);
	// Copy the output in GPU memory back to CPU memory
	clEnqueueReadBuffer(CommandQueue, GPUOutputVector, CL_TRUE, 0, sizeof(HostOutputVector[0]) * MYSIZE, @HostOutputVector[0], 0, nil, nil);
  // Free memory
	clReleaseMemObject(GPUVector1);
	clReleaseMemObject(GPUVector2);
	clReleaseMemObject(GPUOutputVector);
	clReleaseCommandQueue(CommandQueue);
	clReleaseKernel(OpenCLVectorAdd);
  clReleaseProgram(OpenCLProgram);
 	clReleaseContext(context);
  // List results
  ListOutput.Clear;
  for i:=0 to MYSIZE-1 do ListOutput.Items.Add(Format('%g + %g = %g', [HostVector1[i], HostVector2[i], HostOutputVector[i]]));
  Cursor:=crDefault;
end;

procedure TFormMain.BtnFncNotLoadedClick(Sender: TObject);
begin
  ShowMessage(OpenCL_func_not_loaded_str);
end;

end.
